﻿local _KEY,_NAME,_VER = "JD","MultiPanel","0 (alpha)" --or StackPanels?
local minfarbuild = 3892 -- requires Far 3.0 build >=3892 (hDlg:send)

local F = far.Flags
local build = select(4,far.AdvControl(F.ACTL_GETFARMANAGERVERSION,true)) --3,0,0,3892,0
if build<minfarbuild
   and 2~=far.Message(("FAR 3 build %s or newer required.\nYour build: %s")
                      :format(minfarbuild,build),_NAME,"Cancel;Continue","lw")
     then return
end

------------------------------------------------
--Settings

local menuDelayDefault = 200
local S = mf.mload(_KEY,_NAME) or { -- default settings:
}

local macroKey = "`"         -- Ctrl~ / CtrlShift~

-------------------------------------------------------------
--- some general definitions
local function LockState(xLock,State) --set State: 0|1|2=toggle
                                      --get State: nil|-1
  local Lock = {Num=0, Caps=1, Scroll=2}
  local modeIsOn,keyIsPressed = 1,0xff80
  if State then win.Sleep(80) else State=-1 end
  local r = mf.flock(assert(Lock[xLock]),State)
  return band(r,modeIsOn)==modeIsOn,
         band(r,keyIsPressed)==keyIsPressed
end

--[[
RAlt,LAlt  = 0x1,0x2
RCtrl,LCtrl = 0x4,0x8
Shift = 0x10
NumLock,ScrollLock,CapsLock = 0x20,0x40,0x80
ENHANCED_KEY = 0x100
--]]
local RCtrl,LCtrl = 0x4,0x8            --State after last mf.waitkey!!
local function ModState(Mod) return band(Mouse.LastCtrlState,Mod) end

--------------------------------------------------------
-- MultiPanel bare logic
local ACTIVE,PASSIVE = 1,0
local function getSelection(p,Info)
  if Info.SelectedItemsNumber==1 then
    local Item = panel.GetSelectedPanelItem(nil,p,1)
    return band(Item.Flags,F.PPIF_SELECTED)~=0 and {Item.FileName}
  elseif Info.SelectedItemsNumber>0 then
    local Sel = {}
    for i=1,Info.SelectedItemsNumber do
      Sel[i] = panel.GetSelectedPanelItem(nil,p,i).FileName
    end
    return Sel
  end
end

local function setSelection(p,Sel)
  if not Sel then return end
  local Info = panel.GetPanelInfo(nil,p)
  local idx = {}
  for i=1,Info.ItemsNumber do
    local FileName = panel.GetPanelItem(nil,p,i).FileName
    for j=1,#Sel do
      if Sel[j]==FileName then idx[#idx+1]=i end
    end
  end
  panel.SetSelection(nil,p,idx,true)
end

local function getPanel(p)
  local Info = panel.GetPanelInfo(nil,p)
  local info = {}
  if band(Info.Flags,F.PFLAGS_PLUGIN)==F.PFLAGS_PLUGIN then
    info = {
      isPlugin=true,
      notShortcut=band(Info.Flags,F.PFLAGS_SHORTCUT)~=F.PFLAGS_SHORTCUT,
      Format=panel.GetPanelFormat(nil,p),
      Prefix=panel.GetPanelPrefix(nil,p),
    }
  end
  info.Info=Info
  info.Sel=getSelection(p,Info)
  info.Directory=panel.GetPanelDirectory(nil,p)
  return info
end

local function setPanel(p,info)
  if not info then return end --debug
  panel.SetPanelDirectory(nil,p,info.Directory)
  local Info = info.Info
  local Flags = Info.Flags
  panel.SetSortMode(nil,p,Info.SortMode)
  panel.SetNumericSort(nil,p,band(Flags,F.PFLAGS_NUMERICSORT)~=0)
  panel.SetSortOrder(nil,p,band(Flags,F.PFLAGS_REVERSESORTORDER)~=0)
  panel.SetViewMode(nil,p,Info.ViewMode)
  setSelection(p,info.Sel)
  panel.RedrawPanel(nil,p,info.Info)
end

local Panels = {}
function Panels:Push(p,info)                   -- add new
  info = info or getPanel(p)
  if info.notShortcut then
    panel.ClosePanel(p==PASSIVE and F.PANEL_PASSIVE or F.PANEL_ACTIVE)
    panel.RedrawPanel(nil,p)
    info = getPanel(p)
  end
  table.insert(self,1,info)
end
function Panels:Clone(i,info)                  -- add clone
  info = info or self[i]
  --if info.notShortcut then return end --debug
  table.insert(self,i,info)
end
function Panels:Remove(i)                      -- remove
  return table.remove(self,i)
end
function Panels:Close(p,info)                  -- close!!
  if #self>0 then
    self:Pop(p,1)
  else
    info = info or getPanel(p)
    if info.isPlugin then
      panel.ClosePanel(p==PASSIVE and -2 or nil)
      panel.RedrawPanel(nil,p)
    end  
  end
end
--function Panels:Top() return self[1] end       -- most recent Panel
function Panels:Pop(p,i)                       -- pop
  --self:Push(p)
  setPanel(p,self:Remove(i))
end

local Panels = setmetatable(S.Panels or {},{__index=Panels})

--load/save
local function saveSettings()
  S.Panels=Panels
  mf.msave(_KEY,_NAME,S)
end

Event { description="[MultiPanel] Save settings (on exit)";
  group="ExitFAR";
  action=saveSettings;
}

local MPid = "D3656A7D-A75F-40F1-AB9B-59D962CE2224"
local function isMultiPanelMenu() return Menu.Id==MPid end;

NoEvent{ description="[MultiPanel] (auto-switching)"; group="DialogEvent";
  condition=function(Event,Param)
    return 
       Event==F.DE_DLGPROCINIT and Param.Msg==F.DN_LISTCHANGE --bug: http://bugs.farmanager.com/view.php?id=1301#c12736
       and isMultiPanelMenu()
  end;
  action=function(Event,Param)
    mf.postmacro(Keys,"Enter")
    --Param.hDlg:send(F.DM_CLOSE) -- this would cause flicker!!
  end;
}

Event{ description="[MultiPanel] (close on release)"; group="DialogEvent";
  condition=function(Event,Param)
    return Event==F.DE_DLGPROCINIT and Param.Msg==F.DN_ENTERIDLE
       and isMultiPanelMenu() and ModState(LCtrl+RCtrl)==0
  end;
  action=function(Event,Param)
    mf.postmacro(Keys,"Esc")
  end;
}

------------------------------------------------
-- MultiPanel-menu macros
Macro { description="[MultiPanel] Prev/Next";
  area="Menu"; key="/^[LR]Ctrl(Shift)?(CapsLock|Tab|`)$/";
  condition=isMultiPanelMenu;
  action=function() Keys(akey(1):find"Shift" and "Up" or "Down") end;
}

Macro { description="[MultiPanel] Prev/Next (additional)";
  area="Menu"; key="/^[LR]Ctrl(Up|Down|PgUp|PgDn|Num9|Num3)$/";
  condition=isMultiPanelMenu;
  action=function() Keys(akey(1):match"^R?Ctrl(.+)$") end;
}

Macro { description="[MultiPanel] Scroll long titles";
  area="Menu"; key="CtrlRight CtrlLeft";
  condition=isMultiPanelMenu;
  action=function() Keys("Alt"..akey(1):match"Ctrl(.+)") end;
}

Macro { description="[MultiPanel] Goto [0-9A-Z]";
  area="Menu"; key="/^[RL]Ctrl\\w$/";
  condition=isMultiPanelMenu;
  action=function() Keys(akey(1):sub(-1)) end;
}

local TypeW = 10
--local FARid = win.Uuid"00000000-0000-0000-0000-000000000000"
local function prepareMenuItems()
  local Items = {}
  for i=0,#Panels do
    local info = Panels[i] or getPanel(ACTIVE)
    local hk = i
    hk = hk<=9 and hk or hk<=35 and string.char(("A"):byte()+hk-10)
    hk = hk and "&"..hk.."." or ""
    local Dir = info.Directory
    local Type = not info.isPlugin and "Panel" or info.Prefix:match"^[^:]*"
    Type = Type..(" "):rep(TypeW-Type:len())
    local file = Dir.File:match"[^\\]+$"
    local Name = Dir.Name:match".+" or "[Root]"
    Name = info.isPlugin and ((file and file or info.Format)..":"..Name) 
                          or Name
    text = ("%2s %-$s %s %s"):gsub("%$",TypeW)
                             :format(hk,Type,info.Sel and "*" or " ",Name) --Sel
    Items[i+1] = {
      info=info,
      text=text,
      checked = not info.isPlugin and "≡", -- =≡■
      grayed = info.notShortcut,
      --disable
      --hidden
    }
  end
  return Items
end
local Items
local function new_pos(pos,key)
  if key~="" then
    local d = 
              key:find("Up$") and 1 or
              key:find("Down$") and -1 or
              key:find("^R?CtrlShift") and -1 or
              key:find("^R?Ctrl") and 1
    if d then
      pos = pos+d
      return pos>#Items and 1 or pos<1 and #Items or pos
    end
  end
end

local BreakKeys = { --http://msdn.microsoft.com/library/dd375731
  {BreakKey="ESCAPE",br=true},
  {BreakKey="C+RETURN",   action=function(self,pos)
     Panels:Push(PASSIVE)
     Panels:Pop(PASSIVE,pos)
   end,br = true},
  {BreakKey="C+F3",      action=function(self,pos)
     if pos>1 then
       Panels:Clone(pos,Items[pos].info)
     else
       Panels:Push(ACTIVE,Items[pos].info)
     end
     Items = prepareMenuItems()
   end},
  {BreakKey="C+F4",      action=function(self,pos)
     if pos==1 then
       Panels:Close(ACTIVE,Items[1].info)
     else
       Panels:Remove(pos-1)
     end
     Items = prepareMenuItems()
   end},
  {BreakKey="C+F8",      action=function()
     if far.Message("\nDelete saved settings?\n",_NAME,";OkCancel") then
       mf.mdelete(_KEY,_NAME)
     end
   end},
  {BreakKey="S+F9",      action=saveSettings},
  {BreakKey="C+F1",      action=function(self)
     far.Message([[

See Bottom title for hotkeys

[to be continued...]

]],_NAME.." r".._VER,nil,"kl")
     self.br = ModState(LCtrl+RCtrl)==0
   end},
}
local Props = {Title="MultiPanel",Id=win.Uuid(MPid),Flags=F.FMENU_WRAPMODE}
local Bottom = "F3:Clone, F4:Close, Release:Swap, Enter:Swap with passive"

local function showMultiPanel(key)
  Items = prepareMenuItems()
  local item = Items[1]
  local pos = item.grayed and 1 or new_pos(1,key)
  repeat
    Props.Bottom = Bottom
    Props.SelectIndex = pos
    item,pos = far.Menu(Props, Items, BreakKeys)
    if item.action then
      item:action(pos) --if BreakKey
    else               --if Ctrl released
      if pos~=1 then
        Panels:Push(ACTIVE,Items[1].info)
        Panels:Pop(ACTIVE,pos)-- -1+1
      end
      break
    end
  until item.br
  return pos
end

--todo appearance
--todo Help
--todo Info QView Tree ShellAutoCompletion
--todo PFLAGS_VISIBLE
--todo Extra Settings
--todo local Settings?
--todo autoswitch?

--local allPanels  = "Shell Info QView Tree Search ShellAutoCompletion"
local allPanels  = "Shell Search"
Macro { description="[MultiPanel] switch";
  area=allPanels; key="/^LCtrl(Shift)?("..macroKey..")$/";
  action=function()
    showMultiPanel(akey(1))
    if macroKey=="CapsLock" then LockState("Caps",0) end
  end;
}
Macro { description="[MultiPanel] goto Panels";
  area="Editor Viewer"; key="/^LCtrl(Shift)?("..macroKey..")$/";
  action=function()
    Keys"F12 0"
    if Area.Shell then showMultiPanel "" end
    if macroKey=="CapsLock" then LockState("Caps",0) end
  end;
}
Macro { description="[MultiPanel] store active panel";
  area=allPanels; key="CtrlF3";
  action=function()
    local p = ACTIVE
    local info = getPanel(p)
    if info.notShortcut then
      showMultiPanel""
    else
      Panels:Push(p,info)
    end
  end;
}
Macro { description="[MultiPanel] store passive panel";
  area=allPanels; key="CtrlShiftF3";
  action=function()
    local p = PASSIVE
    local info = getPanel(p)
    if info.notShortcut then
      mf.beep()
    else
      Panels:Push(p,info)
    end
  end;
}
Macro { description="[MultiPanel] restore prev. panel";
  area=allPanels; key="CtrlF4";
  action=function()
    Panels:Close(ACTIVE)
  end;
}
Macro { description="[MultiPanel] restore prev. panel as passive";
  area=allPanels; key="CtrlShiftF4";
  action=function()
    Panels:Close(PASSIVE)
  end;
}
